'use strict'

const debug = require('debug')('bluefyre:action:http_request')
const RequestBlockedError = require('../errors').RequestBlockedError
const is_sqli = require('./sqli').action
const is_nosqli = require('./nosqli').action
const is_xss = require('./xss').action
const utils = require('../util/utils')
const qs = require('querystring')


function checkIpAddress(req) {
  let ip = req && req.connection && req && req.connection.remoteAddress

  debug('Request IP %s', ip)

  if (ip in exports.blocked_ips) {
    return true
  }
}

function checkSqli(req) {
  let details

  if (details = is_sqli(req.body)) {
    throw new RequestBlockedError('Payload not allowed.', 'SQLI', { details , in: 'body', type: 'sqli'})
  }

  if (details = is_sqli(req.query)) {
    throw new RequestBlockedError('Query not allowed.', 'SQLI', { details, in: 'query', type: 'sqli' })
  }

  if (details = is_sqli(req.path)) {
    throw new RequestBlockedError('Path not allowed.', 'SQLI', { details, in: 'path', type: 'sqli' })
  }
}

function checkNosqli(req) {
  let details

  if (details = is_nosqli(req.body)) {
    throw new RequestBlockedError('Payload not allowed.', 'NOSQLI', { details, in: 'body', type: 'nosqli' })
  }

  if (details = is_nosqli(req.query)) {
    throw new RequestBlockedError('Query not allowed.', 'NOSQLI', { details, in: 'query', type: 'nosqli' })
  }

  if (details = is_nosqli(req.path)) {
    throw new RequestBlockedError('Path not allowed.', 'NOSQLI', { details, in: 'path', type: 'nosqli' })
  }
}

function checkXSS(req) {
  let details

  if (details = is_xss(req.body)) {
    throw new RequestBlockedError('Payload not allowed.', 'XSS', { details, in: 'body', type: 'xss' })
  }

  if (details = is_xss(req.query)) {
    throw new RequestBlockedError('Query not allowed.', 'XSS', { details, in: 'query', type: 'xss' })
  }

  if (details = is_xss(req.path)) {
    throw new RequestBlockedError('Path not allowed.', 'XSS', { details, in: 'path', type: 'xss' })
  }
}

exports.action = function (data = {}) {
  const { req } = data

  debug('http_request action received data: %j', req.body)

  try {
    let errBF = new Error()

    if (checkIpAddress(req)) {
      debug('IP address blocked')
      throw new RequestBlockedError('IP address not allowed.', 'IPADDR')
    }

    checkSqli(req)
    checkNosqli(req)
    checkXSS(req)

  } catch (err) {
    if (err instanceof RequestBlockedError) {
      err.details.action = 'http_request'

      const block_config = global.__BLUEFYRE.getConfig().request_blocking

      if (err.code == 'SQLI' && block_config.sqli) {
        throw err
      }

      if (err.code == 'NOSQLI' && block_config.nosqli) {
        throw err
      }

      if (err.code == 'XSS' && block_config.xss) {
        throw err
      }


      err.blocked(false)
      // configuration said not to block but we still report
      __BLUEFYRE.reportIncident({ err, req })
    }

    debug('Action error: %o', err)
  }

  debug('Request cleared.')
  return false
}

exports.blocking = true

exports.blocked_ips = {'something': true}
