// bluefyreagent.js
// author: sabin.thomas
// Main agent interface

'use strict'

const util = require('util')
const Receiver = require('./receiver.js')
const EventEmitter = require('events').EventEmitter
const debug = require('debug')('bluefyre:agent')
const readFilesIn = require('./util/read-files-in')
const crypto = require('crypto')
const os = require('os')
const Event = require('./models/Event')
const EnvironmentInfoEvent = require('./models/EnvironmentInfoEvent')
const DependenciesEvent = require('./models/DependenciesEvent')
const Attack = require('./models/AttackEvent')
const namespace = require('./util/namespace')
const pkgUtil = require('./util/package')
const uuid = require('uuid')
const Processor = require('./processor')
const config = require('./config')

debug('Bluefyre Agent version %s starting.', config.version)

const CORENODELIBS = readFilesIn(__dirname + '/instrumentation/core')
delete CORENODELIBS.instrumented_modules

const COREJSLIBS = readFilesIn(__dirname + '/instrumentation/js')

debug('Instrumented node core libraries:', Object.keys(CORENODELIBS).join(', '))
debug('Instrumented Javascript native libraries:', Object.keys(COREJSLIBS).join(', '))

function BluefyreAgent() {
  EventEmitter.call(this)
  this.receiver = new Receiver(this)
  this.processor = new Processor()
}

util.inherits(BluefyreAgent, EventEmitter)

BluefyreAgent.VERSION = config.version

BluefyreAgent.prototype.loadCore = function loadCore() {
  var agent = this

  Object.keys(CORENODELIBS).forEach(function cb(item) {
    debug('loading', item)
    require('./instrumentation/core/' + CORENODELIBS[item])(agent)
  })
}

BluefyreAgent.prototype.loadCoreJS = function loadCoreJS() {
  var agent = this

  Object.keys(COREJSLIBS).forEach(function cb(item) {
    debug('loading', item)
    require('./instrumentation/js/' + COREJSLIBS[item])(agent)
  })
}

BluefyreAgent.prototype.loadModule = function loadModule() {
  debug('Hook into module load')
  var interimMod = require('module')
  require('./instrumentation/core/module')(this, interimMod)
}

BluefyreAgent.prototype.loadNativeJS = function loadNativeJS() {
  debug('changing nativeJS')
  require('./instrumentation/native/index')(this)
}

BluefyreAgent.prototype.start = function start(cb) {
  if (!cb) {
    throw new TypeError('no callback')
  }

  this.receiver.start()
  this.loadModule()
  this.loadCore()
  this.loadCoreJS()
  this.loadNativeJS()
  this.agentInstanceId()
  this._sendEnvInfo()
  this._sendDependencyInfo()

  process.nextTick(cb)
}

BluefyreAgent.prototype.setEvent = function setEvent(payload) {
  this.receiver.addEvent(payload)
}

BluefyreAgent.prototype.agentInstanceId = function() {
  if (this._agent_instance_id) {
    return this._agent_instance_id
  }

  this._agent_instance_id = uuid.v4()

  return this._agent_instance_id
}

/**
 * Capture various environment data that we'll send once
 * at the start of this agent instance.
 */
BluefyreAgent.prototype._captureEnvironment = function() {
  this._env = Object.assign({
    hostname: os.hostname(),
    versions: process.versions,
    os: {
      platform: os.platform(),
      arch: os.arch(),
      release: os.release(),
      type: os.type()
    },
    network: this._captureNetworkEnvironment()
  }, this.getAgentInfo())
}

BluefyreAgent.prototype.getAgentInfo = function () {
  return {
    agent_version: config.version,
    agent_instance_id: this.agentInstanceId(),
    agent_id: config.agent_id,
    license_key: config.license_key
  }
}

BluefyreAgent.prototype._captureNetworkEnvironment = function() {
  const interfaces = os.networkInterfaces()

  const external = {}

  Object.keys(interfaces).forEach(key => {
    const ext = interfaces[key].filter(iface => !iface.internal)

    if (ext.length) {
      external[key] = ext
    }
  })

  return external
}

BluefyreAgent.prototype._captureDependencies = function() {
  
  try{
    pkgUtil.getPackages((dependencyJson) => {
      this._dependencylist = Object.assign(dependencyJson, this.getAgentInfo())  
      //debug('captureDependencies', this._dependencylist)
    })  
  } catch(err){
    debug('could not capture dependencies, setting dependencies to emtpy', err)
    this._dependencylist = Object.assign({}, this.getAgentInfo())  
  }
}

BluefyreAgent.prototype.environment = function() {
  if (this._env) return this._env

  this._captureEnvironment()

  return this.environment()
}

BluefyreAgent.prototype._sendEnvInfo = function() {
  const ev = new EnvironmentInfoEvent(this.environment())
  this.setEvent(ev)
}

BluefyreAgent.prototype.dependencyList = function() {  
  if (this._dependencylist){
    debug('already populated dependencyList previously')
    //debug('already populated dependencyList previously, returning', this._dependencylist)
    return this._dependencylist
  } 

  this._captureDependencies()

  return this.dependencyList()
}

BluefyreAgent.prototype._sendDependencyInfo = function() {
  //TODO: Change this to be new PackageInfo event
  const ev = new DependenciesEvent(this.dependencyList())
  this.setEvent(ev)
}

BluefyreAgent.prototype.runHook = function (hook_name, module_name, payload, cb) {
  return this.processor.execute(hook_name, module_name, payload, cb)
}

BluefyreAgent.prototype.parseReqInfo = function parseReqInfo(req) {
  const req_id = req.headers['x-request-id'] || req.id || uuid.v4()

  const info = { id: req_id }

  info.url = req.url
  info.headers = req.headers
  info.body = req.body // might be yuge tho!!!!
  info.ip_address = (req.headers['x-forwarded-for'] ? req.headers['x-forwarded-for'].split(',').pop(): null) ||
    ((req.connection && req.connection.remoteAddress) ? req.connection.remoteAddress: null) ||
    ((req.socket && req.socket.remoteAddress) ? req.socket.remoteAddress: null) ||
    ((req.connection && req.connection.socket && req.connection.socket.remoteAddress) ? req.connection.socket.remoteAddress : null)
  info.method = req.method

  req.__bluefyre = req.__bluefyre || {}
  req.__bluefyre.req = info
}

BluefyreAgent.prototype.reportIncident = function (data = {req: {}, err: {}}) {
  const at = new Attack()

  at.attack = data.err
  at.req = data.req.__bluefyre.req

  this.setEvent(at)
}

BluefyreAgent.prototype.getConfig = function () {
  return config
}

module.exports = BluefyreAgent
