// express.js
// author: sabin.thomas

'use strict'
const utils = require('../../../util/utils.js')
const util = require('util')
const debug = require('debug')('bluefyre:3rdparty:express')
const Event = require('../../../models/Event')
const BluefyreStackError = require('../../../errors').BluefyreStackError

class ExpressEvent extends Event {
  constructor(method, payload = {}, trace) {
    super(payload, trace)
    this._data.module_name = 'express'
    this._data.method_name = method
    this._data.event_type = 'instrumentation'
  }
}

const METHODS = ['get', 'post', 'put', 'patch', 'head', 'delete', 'del']

function wrapUse(bluefyreAgent, expressModule) {
  if (!expressModule){
    debug('No expressModule');
    return
  }

  if (!bluefyreAgent) {
    debug('No Bluefyre Agent');
    return
  }

  var app = expressModule.application
  var useOrig = app.use

  utils.wrapMethod(bluefyreAgent, app, 'use', 'express', function (use) {
    return function __BF_express_use_wrap(middleware) {
      const payload = {
        arguments: []
      }

      if (middleware && middleware.name) {
        payload.middleware_name = middleware.name
        debug(`Detected middleware ${middleware.name}`)
      }

      for (let i in arguments) {
        payload.arguments.push(util.inspect(arguments[i]))
      }

      let ev = new ExpressEvent('use', payload, new BluefyreStackError(__BF_express_use_wrap))
      bluefyreAgent.setEvent(ev)

      // this reports wrong file/line
      if (middleware.name == 'session') {
        use.apply(this, arguments)
        return this.use(__BF_session_hook)
      }

      return use.apply(this, arguments)
    }
  })

  METHODS.forEach(method => {
    const original_method = app[method]

    app[method] = function (path/*, middleware, ..., handler*/) {
      if (method == 'get' && arguments.length === 1) {
        return original_method.apply(this, arguments)
      }

      debug(`${method.toUpperCase()} ${path} wrapped.`)

      const args = utils.argsToArray.apply(null, arguments)

      // let's say that if the last route arg is a function
      // and has no more than 3 arguments then it's probably the route handler
      let maybe_handler = args[args.length - 1]

      if (typeof maybe_handler == 'function' && maybe_handler.length < 4) {
        const original_handler = args.pop()

        const wrapped_handler = function wrapped_handler(req, res, next) {
          const payload = {
            headers: req.headers,
            path,
            url: req.url,
            method,
          }

          const ev = new ExpressEvent('route_handler', payload, new BluefyreStackError(wrapped_handler))
          bluefyreAgent.setEvent(ev)

          debug(`${method.toUpperCase()} ${path} called.`)
          return original_handler.apply(original_handler, arguments)
        }

        args.push(wrapped_handler)
      }

      return original_method.apply(this, args)
    }
  })

  utils.wrapMethod(bluefyreAgent, app, 'express', 'listen', function (listen) {
    return function __BF_express_listen() {
      return listen.apply(this, arguments)
    }
  })
}

function __BF_session_hook(req, res, next) {
  debug('Added extra decoration', req.sessionID, req.cookies)
  next()
}

module.exports = function initialize(bluefyreAgent, expressModule) {
  debug('Loading Express wrapper');

  if (!expressModule){
    debug('Did not load express successfully')
    return
  }

  var arrFunc = ['disable','set']

  utils.genericWrap(bluefyreAgent,
    expressModule,
    arrFunc,
    'express',
    expressModule && expressModule.prototype && expressModule['application'])

  wrapUse(bluefyreAgent, expressModule)
}
