// mysql.js
// author: sabin.thomas

'use strict';

const utils = require('../../util/utils.js')
const util = require('util')
const debug = require('debug')('bluefyre:3rdparty:mysql')
const Event = require('../../models/Event')
const namespace = require('../../util/namespace')
const errors = require('../../errors')
const config = require('../../config')
const block_sqli = config.request_blocking.sqli
const BluefyreStackError = require('../../errors').BluefyreStackError

class MySQLEvent extends Event {
  constructor (method, payload = {}, trace) {
    super(payload, trace)
    this._data.module_name = 'node-mysql'
    this._data.method_name = method
    this._data.event_type = 'instrumentation'
  }
}


module.exports = function initialize(bluefyreAgent, mysql, loaded_module) {
  if (!bluefyreAgent) {
    debug('Missing Bluefyre Agent.')
    return mysql
  }

  if (!mysql) {
    debug('Missing mysql module.')
    return mysql
  }

  const orig_createConnection = mysql.createConnection

  mysql.createConnection = function __BF_WRAP_createConnection(config) {
    const config_str = util.inspect(config)

    const ev = new MySQLEvent('createConnection', { config }, new BluefyreStackError(__BF_WRAP_createConnection))
    bluefyreAgent.setEvent(ev)

    const connection = orig_createConnection.apply(mysql, arguments)
    if (!connection.query.name != '__BF_mysql_method_wrap') {
      wrapQuery(bluefyreAgent, connection, 'query')
    }
    return connection
  }

  const orig_createPool = mysql.createPool

  mysql.createPool = function __BF_WRAP__createPool(config) {
    const ev = new MySQLEvent('createPool', { config })
    bluefyreAgent.setEvent(ev)

    const connection = orig_createPool.apply(mysql, arguments)
    if (!connection.query.name != '__BF_mysql_method_wrap') {
      wrapQuery(bluefyreAgent, connection, 'query')
    }
    return connection
  }

  const orig_createPoolCluster = mysql.createPoolCluster

  mysql.createPoolCluster = function __BF_WRAP__createPoolCluster(config) {
    const pool_cluster = orig_createPoolCluster.apply(mysql, arguments)
    const orig_of = pool_cluster.of

    const ev = new MySQLEvent('createPoolCluster', { config })
    bluefyreAgent.setEvent(ev)

    pool_cluster.of = function __BF_WRAP__createPoolCluster_of() {
      const namespace = orig_of.apply(pool_cluster, arguments)

      if (namespace.query.name != '__BF_mysql_method_wrap') {
        wrapQuery(bluefyreAgent, namespace, 'query')
      }

      return namespace
    }

    const orig_getConnection = pool_cluster.getConnection

    pool_cluster.getConnection = function __BF_WRAP__createPoolCluster_getConnection() {
      const args = utils.argsToArray.apply(null, arguments)
      const orig_cb = args[args.length - 1]

      args[args.length - 1] = function __BF_WRAP_createPoolCluster_getConnection_cb(err, connection) {
        if (connection.query.name != '__BF_mysql_method_wrap') {
          wrapQuery(bluefyreAgent, connection, 'query')
        }

        orig_cb(err, connection)
      }

      orig_getConnection.apply(pool_cluster, args)
    }

    return pool_cluster
  }

  debug('Instrumented MySQL module.')

  return mysql
}

function wrapQuery(agent, connection, method) {
  if (!connection) {
    return debug('No connection to wrap.')
  }

  if (!method) {
    return debug('No method to wrap.')
  }

  const original = connection[method]

  connection[method] = function __BF_mysql_method_wrap(_sql, _values, _cb) {
    const self = this

    let sql = 'UNKNOWN', values = []

    if (typeof _sql === 'function') {
      sql = ''
      values = []
    } else if (typeof _sql === 'object') {
      sql = _sql.sql
      values = _sql.values

      if (typeof _values == 'function') {
        values = undefined
      } else if (_values) {
        values = _values
      }
    } else {
      if (typeof _sql == 'string') {
        sql = _sql
      }

      // could be an array, could be an object
      if (typeof _values != 'function' && _values) {
        values = _values
      }
    }

    const domain = namespace.get('domain')

    let callback = function __BF_sql_hook_cb(err, block) {
      const ev = new MySQLEvent('createQuery', { sql, values }, new BluefyreStackError(__BF_mysql_method_wrap))
      agent.setEvent(ev)

      debug('Query created: %s', sql, namespace.get('req_id'))

      if (block) {
        if (block_sqli) {
          throw new errors.RequestBlockedError('SQLi detected')
        } else {
          debug('Not blocking detected SQLi .', block)
        }
      } else {
        debug('No SQLi detected.')
      }

      return original.call(self, _sql, _values, _cb)
    }

    if (domain) callback = domain.bind(callback)

    return agent.runHook('sqli', 'mysql', {
      sql, values
    }, callback)
  }
}
