// pg.js
// pg module
// author: sabin.thomas

'use strict';

const utils = require('../../util/utils.js')
const util = require('util')
const debug = require('debug')('bluefyre:3rdparty:pg')
const Event = require('../../models/Event')
const namespace = require('../../util/namespace')
const errors = require('../../errors')
const config = require('../../config')
const block_sqli = config.request_blocking.sqli
const BluefyreStackError = require('../../errors').BluefyreStackError

class PostgresEvent extends Event {
  constructor (method, payload = {}, trace) {
    super(payload, trace)
    this._data.module_name = 'pg'
    this._data.method_name = method
    this._data.event_type = 'instrumentation'
  }
}


module.exports = function initialize(bluefyreAgent, pg, loaded_module) {
  if (!bluefyreAgent) {
    debug('Missing Bluefyre Agent.')
    return pg
  }

  if (!pg) {
    debug('Missing pg module.')
    return pg
  }

  const orig_connect = pg.connect
  pg.connect = function __BF_WRAP_connect(config) {
    const config_str = util.inspect(config)

    const ev = new PostgresEvent('connect', {config}, new BluefyreStackError(__BF_WRAP_connect))
    bluefyreAgent.setEvent(ev)

    const connection = orig_connect.apply(pg, arguments)
    return connection
  }

  const Client = pg.Client
  wrapQuery(bluefyreAgent, Client.prototype, 'query')


  debug('Instrumented Postgres module.')

  return pg
}

function wrapQuery(agent, connection, method) {

  if (!connection) {
    return debug('No connection to wrap.')
  }

  if (!method) {
    return debug('No method to wrap.')
  }

  const original = connection[method]

  connection[method] = function wrap(_sql, _values, _cb) {
    const self = this

    let sql = 'UNKNOWN', values = []

    if (typeof _sql === 'function') {
      sql = ''
      values = []
    } else if (typeof _sql === 'object') {
      sql = _sql.sql
      values = _sql.values

      if (typeof _values == 'function') {
        values = undefined
      } else if (_values) {
        values = _values
      }
    } else {
      if (typeof _sql == 'string') {
        sql = _sql
      }

      // could be an array, could be an object
      if (typeof _values != 'function' && _values) {
        values = _values
      }
    }

    const ev = new PostgresEvent('query', {sql, values}, new BluefyreStackError(wrap))
    agent.setEvent(ev)

    debug('Query created: %s', sql, namespace.get('req_id'))

    const domain = namespace.get('domain')

    let callback = function __BF_sql_hook_cb(err, block) {

      if (block) {
        if (block_sqli) {
          throw new errors.RequestBlockedError('SQLi detected')
        } else {
          debug('Not blocking detected SQLi .', block)
        }
      } else {
        debug('No SQLi detected.')
      }

      return original.call(self, _sql, _values, _cb)
    }

    if (domain) callback = domain.bind(callback)

    return agent.runHook('sqli', 'pg', {
      sql, values
    }, callback)
  }
}
