const fork = require('child_process').fork
const actions = require('./actions')
const namespace = require('./util/namespace')
const debug = require('debug')('bluefyre:processor')

class Processor {
  constructor() {
    this._worker_restart_count = 0
    this.worker.on('message', this._response.bind(this))
    this._heartBeatTimer = this.startHeartbeat()

    if (this._heartBeatTimer && this._heartBeatTimer.unref) {
      this._heartBeatTimer.unref()
    }

    this._job_queue = new Map
  }

  startHeartbeat() {
    return setInterval(_ => {
      this.worker.send({heartbeat: true})
    }, 2000)
  }

  get worker() {
    if (this._worker) return this._worker
    return this._initWorker()
  }

  _initWorker() {
    // fix this awful thing tho
    if (process.env.BF_UNIT_TEST) {
      this._worker = new (require('events')).EventEmitter()
      this._worker.send = () => { }
    }

    if (this._worker) {
      return this._worker
    } else {
      this._worker = fork(`${__dirname}/worker`)

      this._worker.on('disconnect', () => {
        this._worker_restart_count++
        this._worker = null
      })

      this._worker.unref()
    }

    return this._worker
  }

  set worker(w) {
    this._worker = w
  }

  execute(action, module_name, payload, cb) {
    if (!this.isValidAction(action)) {
      debug('Invalid action %s', action)
      return
    }

    if (this.isBlocking(action)) {
      return this._executeBlocking(action, module_name, payload, cb)
    } else {
      this._executeAsync(action, module_name, payload, cb)
    }
  }

  _executeBlocking(action_name, module_name, payload, cb) {
    debug(`Executing blocking action ${action_name}`)

    const result = actions[action_name].action(payload)

    debug(`Executed blocking action ${action_name}`)

    if (cb) {
      return cb(null, result)
    }

    return result
  }

  _executeAsync(action, module_name, payload, cb) {
    debug(`Queueing async action ${action}`)
    const req_id = namespace.get('req_id')

    if (!req_id) {
      debug(`PROCESSOR MISSING REQUEST ID, HOOK WILL NOT RUN`, action, module_name, payload)
      return cb()
    }

    debug(`req id ${req_id}`)
    this._job_queue.set(req_id, cb)

    this.worker.send({
      action, module_name, payload, req_id
    })
  }

  _response(msg) {
    debug('WORKER RESPONSE', arguments)

    if (msg.req_id) {
      const cb = this._job_queue.get(msg.req_id)
      this._job_queue.delete(msg.req_id)

      if (!cb) {
        debug(`NO CALLBACK FOR REQ ID ${msg.req_id}`)
        return
      }

      cb(null, msg.block)
    } else {
      debug('%%%%%%%%%%%%%%%% WORKER RESPONSE NO REQ ID %%%%%%%', msg)
    }
  }

  isBlocking(action) {
    return action &&
      actions[action] &&
      actions[action].blocking
  }

  isValidAction(action) {
    return action && actions[action]
  }
}

module.exports = Processor