// receiver.js
// author: sabin.thomas

'use strict'

const request = require('request')
const debug = require('debug')('bluefyre:receiver')
const debug_config = require('debug')('bluefyre:config')
const crypto = require('crypto')
const config = require('./config')
const EnvironmentInfoEvent = require('./models/EnvironmentInfoEvent')
const async = require('async')
const namespace = require('./util/namespace')

function Receiver(agent) {
  this._event_bucket = []
  this._agent = agent
}

Receiver.prototype.start = function start() {
  this._send_interval = setInterval(() => {
    this.sendMetrics()
  }, 5000)

  if (this._send_interval && this._send_interval.unref) {
    this._send_interval.unref()
  }
}

Receiver.prototype.shutdown = function shutdown() {
  clearInterval(this._send_interval)
  this._send_interval = null
}

Object.defineProperty(Receiver.prototype, '_timer', {
  get: function () {
    return this._send_interval
  }
})

Receiver.prototype.addEvent = function addEvent(payload) {
  this._event_bucket.push(payload)

  // immediately send to avoid losing this event
  if (payload instanceof EnvironmentInfoEvent) {
    this.sendMetrics()
  }
}


/**
 * See if this can be removed
 */
Receiver.prototype.serialize = function serialize(payload, cb) {
  try {
    var res = JSON.stringify(payload, replacer)
  } catch (err) {
    // capture and send the error
    debug(err, "Unable to serialize payload for method")

    return process.nextTick(() => {
      cb(err)
    })
  }

  return cb(null, res);
}

Receiver.prototype.getHeaders = function () {
  if (!this._agent_info) {
    const info = this._agent.getAgentInfo()

    this._agent_info = {}

    Object.keys(info).forEach(key => {
      const header = key.toLowerCase().replace(/[_:]/g, '-')
      this._agent_info[`bluefyre-${header}`] = info[key]
    })
  }

  debug('namespace.get("req").__bluefyre', namespace.get('req') && namespace.get('req').__bluefyre)

  return this._agent_info
}

Receiver.prototype.postData = function postData(data, cb) {
  request({
    method: 'post',
    uri: `${config.collector_url}/api/events`,
    body: data,
    timeout: config.collector_timeout_ms,
    headers: this.getHeaders()
  }, (err, res, body) => {
    if (err) {
      debug('Failed to send metrics: %o', err)
      return cb(err)
    }

    if (res && res.statusCode != 200) {
      debug(`Unexpected response from collector: ${res.statusCode} ${res.text}`)
      return cb(new Error(`Unexpected response from collector: ${res.statusCode} ${res.text}`))
    }

    debug('Metrics sent.')
    cb()
  })
}

Receiver.prototype.queuedEvents = function queuedEvents() {
  return this._event_bucket.slice(0,config.event_queue_threshold)
}

Receiver.prototype.clearEventBucket = function clearEventBucket(){
  this._event_bucket = [];
}

Receiver.prototype.clearQueuedEvents = function clearQueuedEvents(){
  this._event_bucket = this._event_bucket.slice(config.event_queue_threshold)
}


Receiver.prototype.sendMetrics = function sendMetrics(cb) {
  var eventBucket = this.queuedEvents()

  if (!eventBucket.length) {
    return cb && cb()
  }

  this.clearQueuedEvents()


  debug('eventBucket length pre', eventBucket.length)
  // Split the array into serialize sizes of 50 events each
  const serialize_size = 1000
  const partitionArray = (array, size) => array.map( (e,i) => (i % size === 0) ? array.slice(i, i + size) : null ) .filter( (e) => e )
  eventBucket = partitionArray(eventBucket, serialize_size)
  debug('eventBucket length post', eventBucket.length)

  var self = this
  async.eachSeries(eventBucket, function(item, donecb){
  //eventBucket.forEach(function(item){
    let event_count = item.length
    debug('event_count ', event_count)


     self.serialize(item, (err, serialized) => {
      if (err) {
        // metrics lost here – should we put them back
        // and maybe keep last n metrics in the event bucket
        // when error happens?
        return donecb(err)
      }

       if (serialized) {
        debug ('Attempting to send %d events at %s.', event_count, new Date())

        self.postData(serialized, err => {
          if (err) {
            debug('Failed to send metrics.', err)
            donecb(err)
          } else {
            debug('Sent %d events.', event_count)
            donecb()
          }
        })
      } else {
        donecb()
      }
    })

  }, (err)=>{
    debug('err, after callback', err)
    cb && cb(err)
  })
}


/**
 * We can figure out what to do with functions.
 */
function replacer(key, value) {
  if (typeof value == 'function') return '[Function]'

  if (config.drop_fields.indexOf(key) > -1) {
    debug_config('Dropping field "%s" matching drop list.', key)
    return null
  }

  return value
}

module.exports = Receiver
