//utils.js
//author: sabin.thomas

'use strict'
const util = require('util')
const debug = require('debug')('bluefyre:utils')
const namespace = require('./namespace')
const Event = require('../models/Event')
const _ = require('lodash')
const BluefyreStackError = require('../errors').BluefyreStackError

class GenericEvent extends Event {
  constructor(module_name, method, payload = {}, filePath, lineNum) {
    super(payload)
    this._data.module_name = module_name
    this._data.method_name = method
    this._data.event_type = 'instrumentation'

    this._data.file_path = filePath
    this._data.line_num = lineNum
  }
}

//It is impossible to wrap the constructor, since it is a readonly property in Javscript
// other mechanisms such as refactoring the Node.js module core library to modify the constructor may be possible
// for now, all we do is alert what happened.
exports.genericWrapConstructor = function genericWrapConstructor(bluefyreAgent, inModule, moduleName){
  if (!inModule){
    debug('No module');
  }

  if (!bluefyreAgent){
    debug('No Bluefyre Agent');
  }

  // what's this do
  bluefyreAgent.setEvent(new GenericEvent(moduleName, 'constructor'))
}

//Generic wrapper for all nodejs library modules
exports.genericWrap = function genericWrap(bluefyreAgent, inModule, methods, module_name, sourcePrototype){
  if (!inModule){
    debug('No module');
    return inModule
  }

  if (!bluefyreAgent){
    debug('No Bluefyre Agent');
    return inModule
  }

  if (!methods) {
    debug('No method names');
    return inModule
  }

  if (!Array.isArray(methods)) {
    methods = [methods]
  }

  if (!module_name) module_name = 'UNKNOWN'

  if (!sourcePrototype) {
    debug('No source prototype: %s, %s, %o', methods, module_name, sourcePrototype);
    return inModule
  }

  for (let method_name of methods) {
    exports.wrapMethod(bluefyreAgent, sourcePrototype, method_name, module_name, function wrappedFunc(original_method) {
      return function wrapped_method() {
        const args = exports.argsToArray.apply(this, arguments)

        const payload = { arguments: [] }

        for (let a in args) {
          payload.arguments.push(args[a])
        }

        let errBF = new BluefyreStackError(wrapped_method)
        let filePath = errBF.stack[0].getFileName()
        let lineNum = errBF.stack[0].getLineNumber()

        const ev = new GenericEvent(module_name, method_name, payload, filePath, lineNum)
        bluefyreAgent.setEvent(ev)

        return original_method.apply(this, args)
      }
    })
  }
}


exports.wrapMethod = function wrapMethod(agent, module, method_name, module_name, wrapper) {
  if (!module) {
    debug('No module')
    return module
  }

  if (!agent) {
    debug('No Bluefyre Agent')
    return module
  }

  if (!method_name) {
    debug('No function names')
    return module
  }

  if (!module_name) module_name = '__UNKNOWN__'

  if (!wrapper) {
    debug('No wrapper')
    return module
  }

  if (typeof wrapper != 'function') {
    debug('Wrapper must be a function.')
    return module
  }

  const original_method = module[method_name]

  if (!original_method) {
    debug('Cannot wrap non-existent method %s[%s]', module_name, method_name)
    return
  }

  if (original_method.__BF_original) {
    debug('%s[%s] already wrapped', module_name, method_name)
    return
  }

  let wrapper_method = wrapper(original_method, method_name)

  try {
    module[method_name] = wrapper_method
  } catch (simple_error) {
    try {
      const desc = Object.getOwnPropertyDescriptor(module, method_name)

      Object.defineProperty(module, method_name, {
        enumerable: desc.enumerable,
        configurable: desc.configurable,
        get: wrapper_method,
        set: desc.set
      })
    } catch (define_error) {
      debug('Cannot wrap %s', method_name, define_error)
      return
    }
  }

  try {
    Object.defineProperty(module[method_name], '__BF_original', {
      value: original_method,
      enumerable: false,
      configurable: true,
      writable: false
    })

    Object.defineProperty(module[method_name], '__BF_unwrap', {
      value: function unwrap_method() {
        module[method_name] = original_method
        delete module[method_name].__BF_original
        delete module[method_name].__BF_unwrap
      },
      enumerable: false,
      configurable: true,
      writable: false
    })
  } catch (error) {
    debug('Unable to define properties for method %s', method_name, error)
  }
}


//Obtain public methods
exports.getPublicFuncs = function getPublicFuncs(inObj) {
  if(typeof inObj !== 'object' && typeof inObj !== 'function'){
    debug('Empty Object',inObj)
    return [];
  }

  return Object.keys(inObj).filter(function(k) {
    return /^[a-z]/i.test(k) && typeof inObj[k] === 'function'
  });
}

exports.argsToArray = function argsToArray() {
  var array = new Array(arguments.length)

  for (var i = 0; i < array.length; i++) {
    array[i] = arguments[i]
  }

  return array
}

exports.onceAfter = function (fn) {
  let ret_val

  const timeout = setTimeout(function() {
    debug('Hook timed out, invoking function.')
    fn_once()
  }, 50)

  const fn_once = _.once(function () {
    clearTimeout(timeout)
    ret_val = fn.apply(null, arguments)
  })

  return ret_val
}

exports.checkValue = function checkValue(val, check) {
  if (!val) return false

  if (typeof val == 'string') {
    if (/fail this request/.test(val)) {
      return true
    }

    return check(val)
  } else if (Buffer.isBuffer(val)) {
    let body = val.toString('utf8')
    return check(body)
  } else if (Array.isArray(val)) {
    return _.flatten(val.map(v => {
      return exports.checkValue(v, check)
    }))
  } else if (typeof val == 'object') {
    return _.flatten(exports.walkObject(val, exports.checkValue, check))
  }

  return false
}

exports.walkObject = function walkObject(o, checkValue, check) {
  return Object.keys(o).map(key => {
    debug('WALK OBJ', key, o[key])
    return checkValue(key, check) || checkValue(o[key], check)
  })
}


exports.walkKeys = (o, test) => {
  if (!o || typeof o != 'object') {
    return false
  }

  if (Array.isArray(o)) {
    return _.flatten(o.map(v => exports.walkKeys(v, test)))
  }

  return _.flatten(Object.keys(o).map(k => {
    debug('WALK KEYS', k, o[k])
    return test(k, o[k]) || exports.walkKeys(o[k], test)
  }))
}